/*!
 * @file Option.cc
 * ޥɥ饤󥪥ץβϤӥץݻѿδ
 * (Υե mkoption ˤäƼư줿եǤ)
 */

#include <iostream>        // for cout, endl
#include <cstdlib>         // for atof, atoi
#include <cstdio>          // for sscanf

using std::cout;
using std::cerr;
using std::endl;

#include <getopt.h>        // for getopt_long

#include "Option.h"

char** Option::argv_ = 0;

#define SPACES  "\t\t\t\t"

namespace
{
   const char* VERSION_STRING = "The program version 1.0.0";
   const char* AUTHOR_STRING  = "Copyright (C) Koki Uemura <ayato55@img.tutkie.tut.ac.jp>";
   const int OPTION_HELP = -1;
   const int OPTION_VERSION = -2;
   const int OPTION_F1_NAME = -3;
   const int OPTION_F2_NAME = -4;
   const int OPTION_FILTER_RANGE = -5;
   const int OPTION_WINDOW_RANGE = -6;
   const int OPTION_VAR = -7;
   const int OPTION_RK = -8;
   const int OPTION_HARRIS_TH = -9;
   const int OPTION_PMAX = -10;
   const int OPTION_HARRIS_D_MIN = -11;
   const int OPTION_D_MIN = -12;
   const int OPTION_D_MAX = -13;
   const int OPTION_DIM = -14;
   const int OPTION_RADIUS_DIV = -15;
   const int OPTION_RADIUS_SCALE = -16;
   const int OPTION_ANGLE_DIV = -17;
   const int OPTION_TRIANGLE_LENGTH = -18;
   const int OPTION_PV_TH = -19;
   const int OPTION_STR_TH = -20;
   const int OPTION_TRIANGLE_TH = -21;
   const int OPTION_VOTE_TH = -22;
   const int OPTION_PV_WEIGHT = -23;
   const int OPTION_STR_WEIGHT = -24;
   const int OPTION_TRIANGLE_WEIGHT = -25;
   const int OPTION_FOCAL_LENGTH = -26;
   const int OPTION_DISTORTION = -27;
   const int OPTION_USE_CC = -28;
   const int OPTION_USE_DERIVATIVE = -29;
   const int OPTION_NO_USE_RANSAC = -30;
   const int OPTION_VERBOSE = -31;
}

// ץΥǥեͤɽ
const std::string Option::DEFAULT_F1_NAME("f1");
const std::string Option::DEFAULT_F2_NAME("f2");
const int Option::DEFAULT_FILTER_RANGE = 3;
const int Option::DEFAULT_WINDOW_RANGE = 3;
const double Option::DEFAULT_VAR = 1.0;
const double Option::DEFAULT_RK = 0.04;
const double Option::DEFAULT_HARRIS_TH = 1e-10;
const int Option::DEFAULT_PMAX = 500;
const int Option::DEFAULT_HARRIS_D_MIN = 10;
const double Option::DEFAULT_D_MIN = 10.0;
const double Option::DEFAULT_D_MAX = 60.0;
const int Option::DEFAULT_DIM = 10;
const int Option::DEFAULT_RADIUS_DIV = 3;
const int Option::DEFAULT_RADIUS_SCALE = 2;
const int Option::DEFAULT_ANGLE_DIV = 4;
const int Option::DEFAULT_TRIANGLE_LENGTH = 5;
const double Option::DEFAULT_PV_TH = 10.0;
const double Option::DEFAULT_STR_TH = 20.0;
const double Option::DEFAULT_TRIANGLE_TH = 10.0;
const double Option::DEFAULT_VOTE_TH = 2.0;
const double Option::DEFAULT_PV_WEIGHT = 5;
const double Option::DEFAULT_STR_WEIGHT = 5;
const double Option::DEFAULT_TRIANGLE_WEIGHT = 5;
const double Option::DEFAULT_FOCAL_LENGTH = 600.0;
const double Option::DEFAULT_DISTORTION = 1.0;
const bool Option::DEFAULT_USE_CC = false;
const bool Option::DEFAULT_USE_DERIVATIVE = false;
const bool Option::DEFAULT_NO_USE_RANSAC = false;
const bool Option::DEFAULT_VERBOSE = false;

//! ġؿ
namespace
{
   /*!
    * @a flag ǻꤵ줿ͤбʸɽ
    * @param [in] flag 
    * @retval "Yes" @a flag λ
    * @retval "No" @a flag λ
    */
   const char* GetStringType(bool flag)
   {
      return (flag ? "Yes": "No");
   }
};

// --------------------------------------------------

Option::Option() :
      required_parameter_(2),
      f1_name_(DEFAULT_F1_NAME), is_set_f1_name_(false),
      f2_name_(DEFAULT_F2_NAME), is_set_f2_name_(false),
      filter_range_(DEFAULT_FILTER_RANGE), is_set_filter_range_(false),
      window_range_(DEFAULT_WINDOW_RANGE), is_set_window_range_(false),
      var_(DEFAULT_VAR), is_set_var_(false),
      rk_(DEFAULT_RK), is_set_rk_(false),
      harris_th_(DEFAULT_HARRIS_TH), is_set_harris_th_(false),
      Pmax_(DEFAULT_PMAX), is_set_Pmax_(false),
      harris_d_min_(DEFAULT_HARRIS_D_MIN), is_set_harris_d_min_(false),
      d_min_(DEFAULT_D_MIN), is_set_d_min_(false),
      d_max_(DEFAULT_D_MAX), is_set_d_max_(false),
      dim_(DEFAULT_DIM), is_set_dim_(false),
      radius_div_(DEFAULT_RADIUS_DIV), is_set_radius_div_(false),
      radius_scale_(DEFAULT_RADIUS_SCALE), is_set_radius_scale_(false),
      angle_div_(DEFAULT_ANGLE_DIV), is_set_angle_div_(false),
      triangle_length_(DEFAULT_TRIANGLE_LENGTH), is_set_triangle_length_(false),
      pv_th_(DEFAULT_PV_TH), is_set_pv_th_(false),
      str_th_(DEFAULT_STR_TH), is_set_str_th_(false),
      triangle_th_(DEFAULT_TRIANGLE_TH), is_set_triangle_th_(false),
      vote_th_(DEFAULT_VOTE_TH), is_set_vote_th_(false),
      pv_weight_(DEFAULT_PV_WEIGHT), is_set_pv_weight_(false),
      str_weight_(DEFAULT_STR_WEIGHT), is_set_str_weight_(false),
      triangle_weight_(DEFAULT_TRIANGLE_WEIGHT), is_set_triangle_weight_(false),
      focal_length_(DEFAULT_FOCAL_LENGTH), is_set_focal_length_(false),
      distortion_(DEFAULT_DISTORTION), is_set_distortion_(false),
      use_CC_(DEFAULT_USE_CC), is_set_use_CC_(false),
      use_Derivative_(DEFAULT_USE_DERIVATIVE), is_set_use_Derivative_(false),
      no_use_RANSAC_(DEFAULT_NO_USE_RANSAC), is_set_no_use_RANSAC_(false),
      verbose_(DEFAULT_VERBOSE), is_set_verbose_(false)
{
}

// --------------------------------------------------

void Option::Usage(const char* progname)
{
   cerr << "Usage:"       << "\n"
        << "  " << progname << " " << "[options]"
        << " <file1> <file2>\n\n";

   cerr << "Arguments:" << "\n";
   cerr << "  <file1>" << "\t\t\t\t1st image filename (ppm or pgm).\n";
   cerr << "  <file2>" << "\t\t\t\t2nd image filename (ppm or pgm).\n\n";
   cerr << "Options:" << "\n";

   cerr << "  -1, --feature1=STRING\t\t"
        << "filename of the feature points of the 1st image\n\t\t\t\t"
        << "[" << DEFAULT_F1_NAME << "].\n";
   cerr << "  -2, --feature2=STRING\t\t"
        << "filename of the feature points of the 2nd image\n\t\t\t\t"
        << "[" << DEFAULT_F2_NAME << "].\n";
   cerr << "      --filter_range=INTEGER\t"
        << "size of Gauss filter ["
        << DEFAULT_FILTER_RANGE << "].\n";
   cerr << "      --window_range=INTEGER\t"
        << "window size for computing covariance matrices\n\t\t\t\t"
        << "by differentiation ["<< DEFAULT_WINDOW_RANGE << "].\n";
   cerr << "      --var=VALUE\t\t"
        << "variance of Gauss filter [" << DEFAULT_VAR << "].\n";
   cerr << "      --rk=VALUE\t\t"
        << "parameter of Harris operator ["
        << DEFAULT_RK << "].\n";
   cerr << "      --harris_th=VALUE\t\t"
        << "threshold of Harris value ["
        << DEFAULT_HARRIS_TH << "].\n";
   cerr << "      --PointMax=INTEGER"
        << "\tmaximum number for detecting feature points\n\t\t\t\t"
        << "[" << DEFAULT_PMAX << "].\n";
   cerr << "      --min_dist_point=INTEGER\n\t\t\t\t"
        << "minimum distance between two feature points \n\t\t\t\t"
        << "[" << DEFAULT_HARRIS_D_MIN << "].\n";
   cerr << "      --d_min=VALUE\t\t"
        << "minimum length to profile vector ["
        << DEFAULT_D_MIN << "].\n";
   cerr << "      --d_max=VALUE\t\t"
        << "maximum length to profile vector ["
        << DEFAULT_D_MAX << "].\n";
   cerr << "  -d, --pv_dim=INTEGER\t\t"
        << "dimension of profile vector ["
        << DEFAULT_DIM << "].\n";
   cerr << "  -r, --radius_div=INTEGER\t"
        << "number of division in the radius of standard\n\t\t\t\t"
        << "region [" << DEFAULT_RADIUS_DIV << "].\n";
   cerr << "      --radius_scale=INTEGER\t"
        << "radius of standard region ["
        << DEFAULT_RADIUS_SCALE << "].\n";
   cerr << "  -a, --angle_div=INTEGER\t"
        << "number of division in angle of standard\n\t\t\t\t"
        << "region [" << DEFAULT_ANGLE_DIV << "].\n";
   cerr << "      --triangle_length=INTEGER\t"
        << "dimension of triangle region vector ["
        << DEFAULT_TRIANGLE_LENGTH << "].\n";
   cerr << "  -p, --pv_th=VALUE\t\t"
        << "threshold in similarities between two profile\n\t\t\t\t"
        << "vectors [" << DEFAULT_PV_TH << "].\n";
   cerr << "  -s, --str_th=VALUE\t\t"
        << "threshold in similarities between two standard\n\t\t\t\t"
        << "region vectors [" << DEFAULT_STR_TH << "].\n";
   cerr << "  -t, --triangle_th=VALUE\t"
        << "threshold in similarities between two triangle\n\t\t\t\t"
        << "region vectors [" << DEFAULT_TRIANGLE_TH << "].\n";
   cerr << "  -v, --vote_th=VALUE\t\t"
        << "threshold in the number of votes ["
        << DEFAULT_VOTE_TH << "].\n";
   cerr << "      --pv_weight=VALUE\t\t"
        << "weight to similarities of the profile vectors\n\t\t\t\t"
        << "[" << DEFAULT_PV_WEIGHT << "].\n";
   cerr << "      --str_weight=VALUE\t"
        << "weight to similarities of the standard region\n\t\t\t\t"
        << "vectors [" << DEFAULT_STR_WEIGHT << "].\n";
   cerr << "      --triangle_weight=VALUE\t"
        << "weight to similarities of triangle region\n\t\t\t\t"
        << "vectors [" << DEFAULT_TRIANGLE_WEIGHT << "].\n";
   cerr << "      --focal_length=VALUE\t"
        << "temporal focal length for fundamental matrix\n\t\t\t\t"
        << "computation [" << DEFAULT_FOCAL_LENGTH << "].\n";
   cerr << "      --distortion=VALUE\t"
        << "error distances in epipolar equation ["
        << DEFAULT_DISTORTION << "].\n";
   cerr << "      --use_CC\t\t\t"
        << "use cross correlation measure for similarities\n\t\t\t\t"
        << "[" << GetStringType(DEFAULT_USE_CC) << "].\n";
   cerr << "      --use_Derivative\t\t"
        << "use differential method for computing covariance\n\t\t\t\t"
        << "matrices [" << GetStringType(DEFAULT_USE_DERIVATIVE) << "].\n";
   cerr << "      --no_use_RANSAC\t\t"
        << "do not use RANSAC ["
        << GetStringType(DEFAULT_NO_USE_RANSAC) << "].\n";
   cerr << "      --verbose\t\t\t"
        << "verbose output ["
        << GetStringType(DEFAULT_VERBOSE) << "].\n";
   cerr << "      --help\t\t\t"
        << "show this help.\n";
   cerr << "      --version\t\t\t"
        << "show version information.\n";
   cerr << endl;
}

// --------------------------------------------------

void Option::Version(const char* progname)
{
   cerr << " * " << VERSION_STRING << "\n"
             << " * " << AUTHOR_STRING << "\n"
             << " * " << progname << " " << "--help" << " to show usage." << endl;
}

// --------------------------------------------------

void Option::Parse(int argc, char *argv[])
{
   int c;
   int optval = 0;
   struct option long_options[] = {
      {"feature1", required_argument, &optval, OPTION_F1_NAME},
      {"feature2", required_argument, &optval, OPTION_F2_NAME},
      {"filter_range", required_argument, &optval, OPTION_FILTER_RANGE},
      {"window_range", required_argument, &optval, OPTION_WINDOW_RANGE},
      {"var", required_argument, &optval, OPTION_VAR},
      {"rk", required_argument, &optval, OPTION_RK},
      {"harris_th", required_argument, &optval, OPTION_HARRIS_TH},
      {"PointMax", required_argument, &optval, OPTION_PMAX},
      {"min_dist_point", required_argument, &optval, OPTION_HARRIS_D_MIN},
      {"d_min", required_argument, &optval, OPTION_D_MIN},
      {"d_max", required_argument, &optval, OPTION_D_MAX},
      {"pv_dim", required_argument, &optval, OPTION_DIM},
      {"radius_div", required_argument, &optval, OPTION_RADIUS_DIV},
      {"radius_scale", required_argument, &optval, OPTION_RADIUS_SCALE},
      {"angle_div", required_argument, &optval, OPTION_ANGLE_DIV},
      {"triangle_length", required_argument, &optval, OPTION_TRIANGLE_LENGTH},
      {"pv_th", required_argument, &optval, OPTION_PV_TH},
      {"str_th", required_argument, &optval, OPTION_STR_TH},
      {"triangle_th", required_argument, &optval, OPTION_TRIANGLE_TH},
      {"vote_th", required_argument, &optval, OPTION_VOTE_TH},
      {"pv_weight", required_argument, &optval, OPTION_PV_WEIGHT},
      {"str_weight", required_argument, &optval, OPTION_STR_WEIGHT},
      {"triangle_weight", required_argument, &optval, OPTION_TRIANGLE_WEIGHT},
      {"focal_length", required_argument, &optval, OPTION_FOCAL_LENGTH},
      {"distortion", required_argument, &optval, OPTION_DISTORTION},
      {"use_CC", no_argument, &optval, OPTION_USE_CC},
      {"use_Derivative", no_argument, &optval, OPTION_USE_DERIVATIVE},
      {"no_use_RANSAC", no_argument, &optval, OPTION_NO_USE_RANSAC},
      {"verbose", no_argument, &optval, OPTION_VERBOSE},
      {"help", 0, &optval, OPTION_HELP},
      {"version", 0, &optval, OPTION_VERSION},
      {0, 0, 0, 0}
   };

   argv_ = argv;

   while ((c = ::getopt_long(argc, argv, "1:2:d:r:a:p:s:t:v:", long_options, 0)) != -1)
   {
      switch (c)
      {
        case 0:
           switch (optval)
           {
             case OPTION_F1_NAME:
                is_set_f1_name_ = true;
                if (optarg != 0)
                {
                   f1_name_ = optarg;
                }
                break;
             case OPTION_F2_NAME:
                is_set_f2_name_ = true;
                if (optarg != 0)
                {
                   f2_name_ = optarg;
                }
                break;
             case OPTION_FILTER_RANGE:
                is_set_filter_range_ = true;
                if (optarg != 0)
                {
                   filter_range_ = atoi(optarg);
                }
                break;
             case OPTION_WINDOW_RANGE:
                is_set_window_range_ = true;
                if (optarg != 0)
                {
                   window_range_ = atoi(optarg);
                }
                break;
             case OPTION_VAR:
                is_set_var_ = true;
                if (optarg != 0)
                {
                   var_ = strtod(optarg, 0);
                }
                break;
             case OPTION_RK:
                is_set_rk_ = true;
                if (optarg != 0)
                {
                   rk_ = strtod(optarg, 0);
                }
                break;
             case OPTION_HARRIS_TH:
                is_set_harris_th_ = true;
                if (optarg != 0)
                {
                   harris_th_ = strtod(optarg, 0);
                }
                break;
             case OPTION_PMAX:
                is_set_Pmax_ = true;
                if (optarg != 0)
                {
                   Pmax_ = atoi(optarg);
                }
                break;
             case OPTION_HARRIS_D_MIN:
                is_set_harris_d_min_ = true;
                if (optarg != 0)
                {
                   harris_d_min_ = atoi(optarg);
                }
                break;
             case OPTION_D_MIN:
                is_set_d_min_ = true;
                if (optarg != 0)
                {
                   d_min_ = strtod(optarg, 0);
                }
                break;
             case OPTION_D_MAX:
                is_set_d_max_ = true;
                if (optarg != 0)
                {
                   d_max_ = strtod(optarg, 0);
                }
                break;
             case OPTION_DIM:
                is_set_dim_ = true;
                if (optarg != 0)
                {
                   dim_ = atoi(optarg);
                }
                break;
             case OPTION_RADIUS_DIV:
                is_set_radius_div_ = true;
                if (optarg != 0)
                {
                   radius_div_ = atoi(optarg);
                }
                break;
             case OPTION_RADIUS_SCALE:
                is_set_radius_scale_ = true;
                if (optarg != 0)
                {
                   radius_scale_ = atoi(optarg);
                }
                break;
             case OPTION_ANGLE_DIV:
                is_set_angle_div_ = true;
                if (optarg != 0)
                {
                   angle_div_ = atoi(optarg);
                }
                break;
             case OPTION_TRIANGLE_LENGTH:
                is_set_triangle_length_ = true;
                if (optarg != 0)
                {
                   triangle_length_ = atoi(optarg);
                }
                break;
             case OPTION_PV_TH:
                is_set_pv_th_ = true;
                if (optarg != 0)
                {
                   pv_th_ = strtod(optarg, 0);
                }
                break;
             case OPTION_STR_TH:
                is_set_str_th_ = true;
                if (optarg != 0)
                {
                   str_th_ = strtod(optarg, 0);
                }
                break;
             case OPTION_TRIANGLE_TH:
                is_set_triangle_th_ = true;
                if (optarg != 0)
                {
                   triangle_th_ = strtod(optarg, 0);
                }
                break;
             case OPTION_VOTE_TH:
                is_set_vote_th_ = true;
                if (optarg != 0)
                {
                   vote_th_ = strtod(optarg, 0);
                }
                break;
             case OPTION_PV_WEIGHT:
                is_set_pv_weight_ = true;
                if (optarg != 0)
                {
                   pv_weight_ = strtod(optarg, 0);
                }
                break;
             case OPTION_STR_WEIGHT:
                is_set_str_weight_ = true;
                if (optarg != 0)
                {
                   str_weight_ = strtod(optarg, 0);
                }
                break;
             case OPTION_TRIANGLE_WEIGHT:
                is_set_triangle_weight_ = true;
                if (optarg != 0)
                {
                   triangle_weight_ = strtod(optarg, 0);
                }
                break;
             case OPTION_FOCAL_LENGTH:
                is_set_focal_length_ = true;
                if (optarg != 0)
                {
                   focal_length_ = strtod(optarg, 0);
                }
                break;
             case OPTION_DISTORTION:
                is_set_distortion_ = true;
                if (optarg != 0)
                {
                   distortion_ = strtod(optarg, 0);
                }
                break;
             case OPTION_USE_CC:
                is_set_use_CC_ = true;
                use_CC_ = !use_CC_;
                break;
             case OPTION_USE_DERIVATIVE:
                is_set_use_Derivative_ = true;
                use_Derivative_ = !use_Derivative_;
                break;
             case OPTION_NO_USE_RANSAC:
                is_set_no_use_RANSAC_ = true;
                no_use_RANSAC_ = !no_use_RANSAC_;
                break;
             case OPTION_VERBOSE:
                is_set_verbose_ = true;
                verbose_ = !verbose_;
                break;
             case OPTION_HELP:
                Usage(argv[0]);
                throw UsageException();
                break;
             case OPTION_VERSION:
                Version(argv[0]);
                throw UsageException();
                break;
           }
           break;
        case '1':
           is_set_f1_name_ = true;
           if (optarg != 0)
           {
              f1_name_ = optarg;
           }
           break;
        case '2':
           is_set_f2_name_ = true;
           if (optarg != 0)
           {
              f2_name_ = optarg;
           }
           break;
        case 'd':
           is_set_dim_ = true;
           if (optarg != 0)
           {
              dim_ = atoi(optarg);
           }
           break;
        case 'r':
           is_set_radius_div_ = true;
           if (optarg != 0)
           {
              radius_div_ = atoi(optarg);
           }
           break;
        case 'a':
           is_set_angle_div_ = true;
           if (optarg != 0)
           {
              angle_div_ = atoi(optarg);
           }
           break;
        case 'p':
           is_set_pv_th_ = true;
           if (optarg != 0)
           {
              pv_th_ = strtod(optarg, 0);
           }
           break;
        case 's':
           is_set_str_th_ = true;
           if (optarg != 0)
           {
              str_th_ = strtod(optarg, 0);
           }
           break;
        case 't':
           is_set_triangle_th_ = true;
           if (optarg != 0)
           {
              triangle_th_ = strtod(optarg, 0);
           }
           break;
        case 'v':
           is_set_vote_th_ = true;
           if (optarg != 0)
           {
              vote_th_ = strtod(optarg, 0);
           }
           break;
        case '?':
        default:
           Usage(argv[0]);
           throw BadOptionException();
      }
   }

   // ɬܥѥ᡼¸ߤ뤫ɤå
   if (argc - optind < required_parameter_)
   {
      // ѥ᡼­ʤ
      cerr << " * ERROR: " << "too few argument." << endl;
      Usage(argv[0]);
      throw FewArgumentException();
   }

   // ɬܥѥ᡼ɤ߽Ф
   imgfn1_ = Argument(1);
   imgfn2_ = Argument(2);

}

// --------------------------------------------------

char* Option::Argument(int index)
{
   return argv_[optind + index - 1];
}

