#include <list>
#include <iostream>

using std::list;
using std::cerr;
using std::endl;

#include "TriVecDescTable.h"
#include "CrsPoint.h"
#include "TriangleVector.h"
#include "Gauss.h"
#include "Vote.h"

//1Ȳ2¤Τ줾Ȥ뻰ȥ٥ȥ뷴ˤơ
//ȥ٥ȥ٤׻٤ɼơ֥ɼ롣
//٤ϺǤ餷Ȥ߹碌Ǥۤ礭ͤȤʤꡢ
//Ͱʲξ0Ͱʾξ0礭ͤȤʤ롣
void 
Vote::vote_TriVector_at_point(const TriVectorArray &alpha,
                              const TriVectorArray &beta,
                              list<Couple> &couple_list,
                              const vector<int> &alpha_index_array,
                              const vector<int> &beta_index_array)
{
   //ȥ٥ȥˤɼ
   Couple pair;
   double similarity;
   int alpha_num = (int)alpha_index_array.size();
   int beta_num = (int)beta_index_array.size();
  
   for(int i=0; i<alpha_num; i++){
      int index_alpha = alpha_index_array[i];
      for(int j=0; j<beta_num; j++){
         int index_beta = beta_index_array[j];
         similarity = TriVector::similarity_SSD(alpha[index_alpha],
                                                beta[index_beta],
                                                pv_th, str_th,
                                                pv_weight, str_weight);
         if(similarity > 0){
            voting_table[index_alpha][index_beta] += similarity;
		
            pair.a = index_alpha;  pair.b = index_beta; pair.sim = similarity;
            couple_list.push_back(pair);
         }
      }
   }
   /*  
   //ȥ٥ȥˤɼ
   for(int i=0; i<num1; i++){
   for(int j=0; j<num2; j++){
   if(!alpha[i].is_true() || !beta[j].is_true() ) continue;
   similarity = TriVector::similarity_SSD(alpha[i],beta[j],
   pv_th, str_th);
   if(similarity > 0){
   voting_table[i][j] += similarity;
		
   pair.a = i;  pair.b = j; pair.sim = similarity;
   couple_list.push_back(pair);
   }
   }
   }*/
}


void 
Vote::vote_Triangle(const TriVectorTable &tv_table1,
                    const TriVectorTable &tv_table2,
                    const FeaturePointTable &feature1,
                    const FeaturePointTable &feature2,
                    const Image & img1, const Image &img2,
                    const Gauss & gauss,
                    list<Couple> &couple_list,
                    int alpha, int beta, int triangle_length)
{
   //ѷΰˤɼ
   for(list<Couple>::iterator it1 = couple_list.begin();
       it1 != couple_list.end();    it1++ )
   {
      list<Couple>::iterator next = it1;
      next++;
      for(list<Couple>::iterator it2 = next;
          it2 != couple_list.end(); it2++ )
      {
         if(!tv_table1[it1->a][it2->a].is_true() || !tv_table2[it1->b][it2->b].is_true())
            continue;
	  
         double similarity = TriVector::similarity_SSD(tv_table1[it1->a][it2->a],
                                                       tv_table2[it1->b][it2->b],
                                                       pv_th, str_th,
                                                       pv_weight, str_weight);
         if(similarity == 0.0) continue;

         TriangleVector tr1(img1, gauss, triangle_length, feature1[alpha],
                            feature1[it1->a], feature1[it2->a]);
         TriangleVector tr2(img2, gauss, triangle_length,feature2[beta],
                            feature2[it1->b], feature2[it2->b]);
	  
         if(!tr1.be || !tr2.be ){
            continue;
         }
		
         similarity = TriangleVector::similarity_SSD( tr1,tr2,triangle_th, triangle_weight);
         if(similarity > 0){
            voting_table[alpha][beta]    += similarity;
            voting_table[it1->a][it1->b] += similarity;
            voting_table[it2->a][it2->b] += similarity;
         }
      }
   }

}

void
Vote::extract_one2one_pair(const FeaturePointTable &feature1,
                           const FeaturePointTable &feature2,
                           double vote_th,
                           list<CORRES_POINT> &crsp_list)
{
   for(int i = 0; i < num1; i++){
      Couple max = {-1,-1,vote_th};
      //Ǻͤ򸡺
      for(int j = 0; j < num2; j++){
         if(max.sim < voting_table[i][j]){
            max.a = i; max.b = j; max.sim = voting_table[i][j];
         }
      }
      //Ǻͤ򸡺
      int row_max=max.a;
      for(int k = 0; k < num1; k++){
         if(max.sim < voting_table[k][max.b]){
            row_max = k;
         }
      }
      if(max.a == row_max && max.a != -1){
         CORRES_POINT crsp;
         crsp.alpha      = feature1[max.a].point;
         crsp.beta       = feature2[max.b].point;
         crsp.similarity = voting_table[max.a][max.b];
         crsp_list.push_back(crsp);
      }
   }
}
  

void
Vote::initialize(double _pv_th, double _str_th, double _triangle_th,
                 double _pv_weight, double _str_weight,
                 double _triangle_weight, int _num1, int _num2)
{
   pv_th           = _pv_th;
   str_th          = _str_th;
   triangle_th     = _triangle_th;
   pv_weight       = _pv_weight;
   str_weight      = _str_weight;
   triangle_weight = _triangle_weight;
   num1            = _num1;
   num2            = _num2;
  
   voting_table = new double *[num1];
   if(!voting_table){
      cerr << "can't get memory. voteing_table"<<endl;
      exit(1);
   }
   for(int i=0; i<num1; i++){
      voting_table[i] = new double[num2];
      if(!voting_table[i]){
         cerr << "can't get memory. voteing_table"<<endl;
         exit(1);
      }
   }
   for(int i=0; i<num1; i++){
      for(int j=0; j<num2; j++){
         voting_table[i][j] = 0;
      }
   }
  
}

// ѷΰȻȥ٥ȥΤ줾٤ɼ롣
// tv_table1, tv_table2 : Ʋ鸡Фȥ٥ȥ
// feature1,  feature2  : Ʋ鸡Фħ
// img1, img2           : ϲ
// gauss                : ʿ경ե륿
// triangle_length      : ѷΰʬ䷸ K
// vote_th              : ɼͤ
// crsp_list            : б̳Ǽꥹ(11б)	
void
Vote::voting(const TriVectorTable &tv_table1,
             const TriVectorTable &tv_table2,
             const FeaturePointTable &feature1,
             const FeaturePointTable &feature2,
             const Image & img1, const Image &img2,
             const Gauss & gauss,int triangle_length,
             double vote_th,
             list<CORRES_POINT> &crsp_list)
{
   list<Couple> couple_list;
   //P_iP_jΤȤ뻰ȥ٥ȥȻѷΰѤơɼԤ
   for(int i=0; i<num1; i++){
      for(int j=0; j<num2; j++){
         vote_TriVector_at_point(tv_table1[i], tv_table2[j], couple_list,
                                 tv_table1.useable_index_table[i],
                                 tv_table2.useable_index_table[j]);
         if(couple_list.size() < 2){//ȥ٥ȥбȤ4̤ΤΤϴ
            couple_list.clear();
            continue;
         }
         vote_Triangle(tv_table1, tv_table2, feature1,feature2, 
                       img1, img2, gauss,
                       couple_list,
                       i,j,triangle_length);
         couple_list.clear();
      }
   }
   extract_one2one_pair(feature1, feature2,vote_th,crsp_list);
   crsp_list.sort();
}
