#include <vector>
#include <iostream>

#include <Mat.h>
#include "ransac.h" 
#include "fundamental_estimation.h"
#include "CrsPair.h"
#include "CrsPoint.h"

using std::vector;
using std::cerr;
using std::endl;

//åԥѴؿCrsPair˹碌ơͤǼ롣
//ޤбɸxyɸϤѴ롣
void
CORRES_POINT_to_CrsPair(list<CORRES_POINT> & crsp_list, CrsPair &cp,
                        int width, int height, double focal_length)
{
   int wc = width / 2, hc = height / 2;  
   int num = (int)crsp_list.size();
   list<CORRES_POINT>::iterator it = crsp_list.begin();

   for (int i=0; i< num; i++)
   {
      cp.pp1[i][0]     = (hc - it->alpha[1]) / focal_length;
      cp.pp1[i][1]     = (it->alpha[0] - wc) / focal_length;
      cp.pp1[i][2]     = 1;
      cp.pp2[i][0]     = (hc - it->beta[1])  / focal_length;
      cp.pp2[i][1]     = (it->beta[0] - wc)  / focal_length;
      cp.pp2[i][2]     = 1;
      cp.cumul_prob[i] = it->similarity;
      it++;
   }
   cp.width = width; cp.height = height; cp.fcslen = focal_length;
   cp.corres_num = num;
}

void
show(list<CORRES_POINT> &crsp_list)
{
   cerr << "show list" << endl;
   list<CORRES_POINT>::iterator it = crsp_list.begin();

   while (it != crsp_list.end())
      cerr << it->alpha << "--" << it->beta << endl;  
}

void
show(VecC *pp1, VecC *pp2, int num)
{
   cerr << "show array" << endl;
   for (int i=0; i< num; i++)
      cerr << pp1[i] << "--" << pp2[i] << endl;
}

//ùƤϤ
bool
apply_epipolar_to_crsp(list<CORRES_POINT> &crsp_list,
                       VecC *pp1, VecC *pp2,
                       Mat &F, double epsilon)
{
   list<CORRES_POINT>::iterator it = crsp_list.begin();
   int i = 0;

   while (it != crsp_list.end())
   {
      if (!check_condition_of_constraint(pp1[i],pp2[i],F,epsilon,0))
         it = crsp_list.erase(it);
      else
         it++;
      i++;
   }

   return true;
}


/////////////////////////////////////////////////////////////////////
//Զ«ɤȽ̤ʤΤб
// 
// F             ᤿ù
// crsp_list     бǡ( & )
// img_width     β        
// img_height    ν
// focal_length  Υ
// dist          Ĥߵ
//
//  true     false
//
////wide baselineʲбդΤȤϡ12ξȹ⤵ɬפˤʤ
bool
fundamental_estimation(list<CORRES_POINT> &crsp_list,
                       int img_width, int img_height, double focal_length,
                       double dist)
{
   Mat F_, F;
   int max_support;  
   double epsilon = 2*dist*dist/ (focal_length * focal_length);
   int crs_num = (int)crsp_list.size();
  
   CrsPair cp;  
   cp.pp1 = new VecC[crs_num];
   cp.pp2 = new VecC[crs_num];
   cp.cumul_prob = new double[crs_num];

   if (!cp.pp1 || !cp.pp2 || !cp.cumul_prob)
   {
      cerr << "can't get memory. CrsPair.pp1, pp2 cumul_prob" << endl;
      exit(1);
   }

   // & ɸѴ & ѥ᡼Ǽ
   CORRES_POINT_to_CrsPair(crsp_list, cp, img_width, img_height, focal_length);

   //8̵ФˤRANSACԤ
   max_support = ransac_fundamental(cp, epsilon, MAXITR, 0, F_);
//   cerr<<"max_support"<<max_support<<"/"<<crs_num<<endl;  
   //ùˤбԤ
   if(max_support < 8)
   {
      cerr << "too few correspondence point that we could find" << endl;
      exit(1);
   }
   apply_epipolar_to_crsp(crsp_list, cp.pp1, cp.pp2, F_, epsilon);
   crs_num = crsp_list.size();
   //ٴù(ϷˡȤ)
   CORRES_POINT_to_CrsPair(crsp_list,cp, img_width, img_height, focal_length);
   calc_fundamental(cp.pp1, cp.pp2,crs_num, F, DEGN, epsilon, MAXITRN);
   
   apply_epipolar_to_crsp(crsp_list, cp.pp1, cp.pp2, F, epsilon);
//   cerr << "crs_num " << crs_num << endl;
  
   delete [] cp.pp1;
   delete [] cp.pp2;
   delete [] cp.cumul_prob;
   return true;
}
