#ifndef _QUATERNION_H_
#define _QUATERNION_H_

#include <Vec.h>
#include <Mat.h>

// ---------------------------------------------------------------------- 
// 4Υ饹
// ---------------------------------------------------------------------- 
class Quaternion
{
 private:
    double s;
    VecR v;
 public:
    Quaternion();
    Quaternion(double s, double v0, double v1, double v2);
    Quaternion(double s, const VecR v);
    Quaternion(double s, const VecC v);
    Quaternion(const Mat &m);
    virtual ~Quaternion();

    // Ѵ
    Quaternion & operator =(const Quaternion &q);
    Mat  getMat();
    VecR & getVecR() { return v; };
    double getAngle() {return acos(s) * 2; };
    
    // 黻1
    friend Quaternion   operator +(const Quaternion &q1, const Quaternion &q2);
    friend Quaternion   operator *(const Quaternion &q1, const Quaternion &q2);
    friend Quaternion   operator *(const Quaternion &q, double d);
    friend Quaternion   operator /(const Quaternion &q, double d);
    friend double operator ,(const Quaternion& q1, const Quaternion& q2);

    // 黻2
    void        setIdentity();
    VecC        rotateVec(const VecC &v);
    void        setRotate(const VecR &axis, double angle);
    void        setRotate(const VecC &axis, double angle);
    Quaternion        inverse();    
    Quaternion        normalize();  
    Quaternion        conjugate();  
    double      norm();       
    //Quaternion        exp();
    //Quaternion        log();
    //double      abs();

    // 
    friend Quaternion slerp(const Quaternion &q1, const Quaternion &q2, double t);
    friend Quaternion squad(const Quaternion &q1, const Quaternion &b1, 
	       const Quaternion &b2, const Quaternion &q2, double t);

    // 
#if __GNUC__ < 3
    friend ostream & operator << (ostream &s, const Quaternion &Q);
#else
    friend std::ostream & operator << (std::ostream &s, const Quaternion &Q);
#endif
};
// ---------------------------------------------------------------------- 

#endif
